import os
from functools import wraps

import rook


def on_lambda():
    return 'LAMBDA_TASK_ROOT' in os.environ


def serverless_rook(f):

    if on_lambda():
        @wraps(f)
        def handler(*args, **kwargs):
            rook.start()

            ret = f(*args, **kwargs)

            rook.flush()
            return ret
    else:
        handler = f

    return handler


def setup_zappa_support():

    if on_lambda():
        try:
            from zappa.middleware import ZappaWSGIMiddleware

            original_handler = ZappaWSGIMiddleware.__call__

            @serverless_rook
            def zappa_hook(self, *args, **kwargs):
                return original_handler(self, *args, **kwargs)

            ZappaWSGIMiddleware.__call__ = zappa_hook

        except ImportError:
            # can happen if its not zappa
            pass
    else:
        pass


try:
    from chalice import Chalice

    class RookoutChalice(Chalice):
        @serverless_rook
        def __call__(self, *args, **kwargs):
            return super(RookoutChalice, self).__call__(*args, **kwargs)


except ImportError:
    pass



import sys
import time
import threading
import types

import six
import os
import inspect
import site
import weakref
import imp

from collections import defaultdict

from rook.services.exts.cloud_debug_python.module_explorer import _GetLineNumbers
from six.moves import builtins

from rook.logger import logger

from rook.config import ImportServiceConfig

from rook.serverless import on_lambda

from rook.exceptions.tool_exceptions import RookDependencyError

_SUPPORTED_PY_FORMATS = ['.py', '.pyc']

_BLACKLISTED_PATHS = []


class CountingImportLock(object):
    """
    When locked, only this thread can import (so all imports are guaranteed to have completed,
                                              and there are no partially initialized modules in sys.modules)
    """
    def __init__(self):
        self._count = 0

    def __enter__(self):
        imp.acquire_lock()
        self._count += 1

    def __exit__(self, *args, **kwargs):
        self._count -= 1
        imp.release_lock()

    def get_recursion_count(self):
        return self._count


class ImportService(object):

    NAME = "Import"

    class Notification(object):
        def __init__(self, module_name, module_filename, include_externals, lineno, callback, removed):
            self.module_name = module_name
            self.module_filename = module_filename
            self.callback = callback
            self.removed = removed
            self.include_externals = include_externals
            self.lineno = lineno

    def __init__(self, bdb_location_service):
        self._bdb_location_service = bdb_location_service

        self._modules = frozenset(sys.modules.values())
        self._path_cache = {}
        self._post_import_notifications = {}

        self._thread = None
        self._quit = False

        self._old_import = None

        external_paths = [sys.exec_prefix]
        if hasattr(site, 'getsitepackages'):
            external_paths = external_paths + site.getsitepackages()

        self.external_paths = [os.path.normcase(os.path.realpath(external_path))
                               for external_path in external_paths]

        if on_lambda():
            _BLACKLISTED_PATHS.extend(['/var/task/flask', '/var/task/chalice'])

        import platform
        if platform.system() != 'Windows' and ImportServiceConfig.USE_IMPORT_HOOK:
            platform = platform.python_implementation().lower()

            def import_hook(*args, **kwargs):
                """
                Declared here to make it obvious that you can't replace the function by monkeypatching -
                the C extension will always reference the function set with SetImportHook.
                """
                __rookout__tracebackhide__ = True
                # Locking the global import lock here extends the section for which it is locked -
                # we don't want the import hook to be called from two threads simultaneously,
                # but the scope in which the import lock is held might only encompass the call to the original
                # __import__. The import lock is re-entrant, so it's OK for us to lock it and then for Python
                # to lock it again.
                with import_hook.import_lock:
                    result = self._old_import(*args, **kwargs)

                    try:
                        # if the recursion count is > 1,
                        # then this call to import_hook is a nested import, which means that the parent
                        # import has not finished executing its code yet, but it's already in sys.modules.
                        # if we evaluate now, we will try to place BPs in the module before it finished executing,
                        # potentially resulting in CodeNotFound.
                        if import_hook.import_lock.get_recursion_count() == 1:
                            self.evaluate_module_list()
                    except:
                        pass

                    return result
            # this must be a reentrant lock - imports can cause other imports
            import_hook.import_lock = CountingImportLock()

            if platform == "cpython":
                try:
                    import native_extensions
                except Exception as e:
                    raise RookDependencyError(e)

                native_extensions.SetImportHook(import_hook)
                # atomic swap
                builtins.__import__, self._old_import = native_extensions.CallImportHookRemovingFrames, builtins.__import__
            elif platform == "pypy":
                import __pypy__
                # atomic swap
                builtins.__import__, self._old_import = __pypy__.hidden_applevel(import_hook), builtins.__import__
            else:
                # assertion should never be reached, singleton.py checks platform support
                raise AssertionError("Unsupported platform")
        else:
            self._thread = threading.Thread(target=self._query_thread,
                                            name=ImportServiceConfig.THREAD_NAME)
            self._thread.daemon = True
            self._thread.start()

    def close(self):
        if self._old_import:
            builtins.__import__ = self._old_import

        if self._thread:
            self._quit = True

            # If threading was monkey patched by gevent waiting on thread will likely throw an exception
            try:
                from gevent.monkey import is_module_patched
                if is_module_patched("threading"):
                    time.sleep(ImportServiceConfig.SYS_MODULES_QUERY_INTERVAL)
            except:
                pass

            self._thread.join()

    def register_post_import_notification(self, aug_id, name, filename, include_externals, lineno, callback, removed):
        # Normalize file path
        if filename:
            filename = os.path.normcase(os.path.normpath(filename))

        notification = self.Notification(name, filename, include_externals, lineno, callback, removed)
        with CountingImportLock():
            # Attempt to satisfy notification using known modules
            for module_object in self._modules:

                # Get module details and check if it matches
                module_filename = self._get_module_path(module_object)

                # If module is not valid, ignore
                if not self._is_valid_module(module_object, module_filename):
                    continue

                if module_filename:
                    if self._does_module_match_notification(module_object.__name__, module_filename, notification, self.external_paths):
                        notification.callback(module_object)
                        return False

        # Register notification for future loads
        self._post_import_notifications[aug_id] = notification

        return True

    def remove_aug(self, aug_id):
        notification = self._post_import_notifications.pop(aug_id, None)
        if notification is None:
            return

        notification.removed()

    def clear_augs(self):
        # This does not require a lock - `notifications` may be added to or removed from in a different thread
        # while we still haven't replaced it with an empty dict(),
        # but `notifications` is just a pointer to the same dict object pointed to by
        # `_post_import_notifications`. Once we do replace `_post_import_notifications`
        # with a new dict, we can be sure that no more notifications will be added.
        # At that point, it's safe to iterate over `notifications` and remove all notifications.
        notifications = self._post_import_notifications
        self._post_import_notifications = dict()
        for notification in six.itervalues(notifications):
            notification.removed()

    def _is_valid_module(self, module_object, module_filename):
        return module_filename and os.path.splitext(module_filename)[1] in _SUPPORTED_PY_FORMATS and \
            module_object and inspect.ismodule(module_object) and hasattr(module_object, '__file__')

    def _query_thread(self):
        self._bdb_location_service.ignore_current_thread()

        while not self._quit:
            try:
                with CountingImportLock():
                    self.evaluate_module_list()
            except:
                if logger:
                    logger.exception("Error while evaluating module list")

            # time can be None if interpreter is in shutdown
            if not time:
                return
            time.sleep(ImportServiceConfig.SYS_MODULES_QUERY_INTERVAL)

    def evaluate_module_list(self):
        try:
            # Nobody is waiting for notifications
            if not self._post_import_notifications:
                return

            # No new modules
            if len(self._modules) == len(sys.modules):
                return

            # Get a fresh list
            modules = frozenset(sys.modules.values())
            # self._modules is only replaced (it's immutable - a frozenset), so there's no need to copy it here,
            # just keep a reference to the current self._modules so we don't start working on a different set
            # mid-loop
            old_modules = self._modules

            # For everybody not in the old list, check notifications
            for module in modules:
                module_filename = self._get_module_path(module)
                if module not in old_modules and self._is_valid_module(module, module_filename):
                    self._notify_of_new_module(module.__name__, module, module_filename)

            # Update the "old" list
            self._modules = modules

        except:
            logger.exception("Exception in ImportService")

    def _notify_of_new_module(self, module_name, module_object, module_filename):
        self._trigger_all_notifications_for_module(module_filename, module_name, module_object)

    def _trigger_all_notifications_for_module(self, module_filename, module_name, module):
        augs_to_remove = set()

        if module_filename:
            for aug_id, notification in six.iteritems(self._post_import_notifications.copy()):
                if self._does_module_match_notification(module_name, module_filename, notification,
                                                        self.external_paths):
                    try:
                        augs_to_remove.add(aug_id)
                        notification.callback(module)
                    except:
                        logger.exception("Error on module load callback")
        for aug_id in augs_to_remove:
            self._post_import_notifications.pop(aug_id, None)

    def _get_module_path(self, module):
        result = self._path_cache.get(module)

        if result is not None:
            return result

        if module:
            try:
                path = inspect.getsourcefile(module)
                if not path:
                    if module.__file__.endswith('.pyc'):
                        path = module.__file__.replace('.pyc', '.py')
            except:
                return None

            if path:
                result = os.path.realpath(os.path.normcase(os.path.abspath(path)))
                self._path_cache[module] = result
                return result

        return None

    @staticmethod
    def is_black_listed_path(path):
        for blacklisted_path in _BLACKLISTED_PATHS:
            if path.startswith(blacklisted_path):
                return True
        return False

    @staticmethod
    def _does_module_match_notification(module_name, module_filename, notification, external_paths):
        if not notification.include_externals:
            for external_path in external_paths:
                if module_filename.startswith(external_path):
                    return False

        if notification.module_filename and ImportService.path_contains_path(module_filename, notification.module_filename) or \
                        notification.module_name and module_name == notification.module_name and \
                        not ImportService.is_black_listed_path(module_filename):
            return True
        else:
            return False

    @staticmethod
    def _does_code_object_match_notification(code_object, notification):
        return notification.lineno in _GetLineNumbers(code_object)

    @staticmethod
    def path_contains_path(full_path, partial_path):
        if full_path.endswith(partial_path):
            return len(full_path) == len(partial_path) or full_path[-len(partial_path)-1] in ('/', '\\')
        else:
            return False



class ToolException(Exception):

    def __init__(self, *args):
        super(ToolException, self).__init__(*args)

    def get_type(self):
        return self.__class__.__name__

    def get_message(self):
        try:
            return self.args[0]
        except IndexError:
            return ""

    def get_parameters(self):
        try:
            return self.args[1]
        except IndexError:
            return ()


class RookException(ToolException):
    pass


class RookCommunicationException(RookException):
    pass


class RookInputException(ToolException):
    pass


class WebHookProcessingException(Exception):
    def __init__(self, status_code, content):
        super(WebHookProcessingException, self).__init__("Problem sending webhook %s" % status_code,
                                                         {
                                                             "status_code": status_code,
                                                             "content": content
                                                         })


class WebHookRetryLevelException(Exception):
    def __init__(self, retry_level, valid_levels):
        super(WebHookRetryLevelException, self).__init__("Invalid retry level for webhook %s" % retry_level,
                                                         {
                                                        "retry_level": retry_level,
                                                        "valid_levels": valid_levels}
                                                         )


class RookInvalidArithmeticPath(RookInputException):
    def __init__(self, configuration, error=None):
        super(RookInvalidArithmeticPath, self).__init__("Invalid arithmetic path configuration. configuration: %s, innerException: %s" % (configuration, error),
                                                        {'configuration': configuration, 'error': error})



class RookAugInvalidKey(RookInputException):
    def __init__(self, key, configuration):
        super(RookAugInvalidKey, self).__init__("Failed to get key %s from configuration %s" % (key, configuration),
                                                {'key': key,
                                                 'configuration': configuration})


class RookObjectNameMissing(RookInputException):
    def __init__(self, configuration):
        super(RookObjectNameMissing, self).__init__("Failed to find object name %s" % configuration, {'configuration': configuration})


class RookUnknownObject(RookInputException):
    def __init__(self, object_name):
        super(RookUnknownObject, self).__init__("Failed to find object %s" % object_name, {'object_name': object_name})


class RookInvalidObjectConfiguration(RookInputException):
    def __init__(self, object_name, object_config):
        super(RookInvalidObjectConfiguration, self).__init__("Failed to build object %s" % object_name,
                                                             {'object_name': object_name,
                                                              'object_config': object_config})


class RookSendToRookoutDisabledException(RookInputException):
    def __init__(self):
        super(RookSendToRookoutDisabledException, self).__init__()


class RookMonitorException(RookException):
    pass


class RookHashFailedException(RookMonitorException):
    def __init__(self, module_name):
        super(RookHashFailedException, self).__init__("Failed to calculate hash %s" % module_name, {'module': module_name})


class RookHashMismatchException(RookMonitorException):
    def __init__(self, filepath, expected, calculated, gitBlob=None):
        super(RookHashMismatchException, self).__init__("File hashes do not match! path: %s, expected: %s, calculated: %s" % (filepath, expected, calculated),
                                                        {'filepath': filepath,
                                                         'expected': expected,
                                                         'calculated': calculated,
                                                         'gitBlob': gitBlob})


class RookHashCalculationFailed(RookMonitorException):
    def __init__(self, filepath, class_name, exc):
        super(RookHashCalculationFailed, self).__init__("Hash calculation failed due to unknown error",
                                                        {'file_name': filepath,
                                                         'class_name': class_name,
                                                         'throwable': exc})


class RookBdbFailedException(RookMonitorException):
    def __init__(self, result):
        super(RookBdbFailedException, self).__init__("Failed to set breakpoint %s" % result, {'result': result})


class RookInvalidPositionException(RookException):
    def __init__(self, filename, line, alternatives):
        super(RookInvalidPositionException, self).__init__("Code position is not breakable: %s:%s" % (filename, line), {
            'filename': filename,
            'line': line,
            'alternatives': alternatives
        })


class RookBdbCodeNotFound(RookMonitorException):
    def __init__(self, filename):
        super(RookBdbCodeNotFound, self).__init__("Failed to find code object %s" % filename, {'filename': filename})


class RookBdbSetBreakpointFailed(RookMonitorException):
    def __init__(self, msg):
        super(RookBdbSetBreakpointFailed, self).__init__("Failed to set breakpoint! %s" % msg)


class RookAttributeNotFound(RookMonitorException):
    def __init__(self, attribute):
        super(RookAttributeNotFound, self).__init__("Failed to get attribute %s" % attribute, {'attribute': attribute})


class RookKeyNotFound(RookMonitorException):
    def __init__(self, key):
        super(RookKeyNotFound, self).__init__("Failed to get key %s" % key, {'key': key})


class RookMethodNotFound(RookMonitorException):
    def __init__(self, namespace_type, method):
        super(RookMethodNotFound, self).__init__("Namespace method not found %s" % method,
                                                 {'namespace': namespace_type.__name__,
                                                  'method': method})


class RookInvalidMethodArguments(RookMonitorException):
    def __init__(self, method, args):
        super(RookInvalidMethodArguments, self).__init__("Bad method arguments: method: %s, args %s" % (method, args),
                                                 {'method': method,
                                                  'arguments': args})


class RookWriteAttributeNotSupported(RookMonitorException):
    def __init__(self, namespace_type, attribute):
        super(RookWriteAttributeNotSupported, self).__init__("Namespace %s does not support write" % namespace_type.__name__,
                                                             {'namespace': namespace_type.__name__,
                                                              'attribute': attribute})


class RookOperationReadOnly(RookMonitorException):
    def __init__(self, operation_type):
        super(RookOperationReadOnly, self).__init__("Operation does not support write: %s" % operation_type.__name__,
                                                    {'operation': operation_type.__name__})


class RookRuleRateLimited(RookException):
    def __init__(self):
        super(RookRuleRateLimited, self).__init__("Breakpoint was disabled due to rate-limiting. "
                                                  "For more information: https://docs.rookout.com/docs/breakpoints-tasks.html#rate-limiting")


class RookRuleMaxExecutionTimeReached(RookException):
    def __init__(self):
        super(RookRuleMaxExecutionTimeReached, self).__init__("Breakpoint was disabled because it has reached its maximum execution time")


class RookNoHttpServiceRegistered(RookException):
    def __init__(self):
        super(RookNoHttpServiceRegistered, self).__init__("No http service registered")


class RookUnsupportedLocation(RookException):
    def __init__(self, location):
        super(RookUnsupportedLocation, self).__init__("Unsupported aug location was specified: %s" % location,
                                                      {'location': location})


class RookInterfaceException(RookException):
    def __init__(self, error_string):
        super(RookInterfaceException, self).__init__(error_string)


class RookVersionNotSupported(RookException):
    def __init__(self, error_string):
        super(RookVersionNotSupported, self).__init__(error_string)


class RookLoadError(RookException):
    def __init__(self, message):
        super(RookLoadError, self).__init__(message)


class RookMissingToken(RookException):
    def __init__(self):
        super(RookMissingToken, self).__init__('No Rookout token was supplied. '
                                               'Make sure to pass the Rookout Token when starting the rook')


class RookInvalidToken(RookException):
    def __init__(self, token):
        super(RookInvalidToken, self).__init__('The Rookout token supplied (%s...) is not valid; '
                                               'please check the token and try again' % token[:6])


class RookServiceMissing(RookException):
    def __init__(self, service):
        super(RookServiceMissing, self).__init__('Rookout service is missing',
                                                 {'service': service})


class RookInvalidOptions(RookException):
    def __init__(self, description):
        super(RookInvalidOptions, self).__init__(description)


class RookInvalidLabel(RookException):
    def __init__(self, label):
        super(RookInvalidLabel, self).__init__("Invalid label: must not start with the '$' character",
                                               {'label': label})


class RookNonPrimitiveObjectType(RookInputException):
    def __init__(self, path):
        super(RookNonPrimitiveObjectType, self).__init__(
            "Object %s must be of primitive type, such as: string, int, long etc" % path,
            {'path': path})


class RookExceptionEvaluationFailed(RookInputException):
    def __init__(self, expression):
        super(RookExceptionEvaluationFailed, self).__init__(
            "Failed to evaluate expression %s" % expression,
            {'expression': expression})


class RookSourceFileNotFound(RookInputException):
    def __init__(self, file_name):
        super(RookSourceFileNotFound, self).__init__(
            "Failed to find file %s" % file_name,
            {'file_name': file_name})


class RookDependencyError(RookException):
    def __init__(self, exc):
        super(RookDependencyError, self).__init__("%s, Make sure to run 'pip install rook' on the same platform/architecture. "
                                                  "For more information: https://docs.rookout.com/docs/sdk-setup.html#building-1" % exc,
                                                        {'throwable': exc})



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: messages.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import variant_pb2 as variant__pb2
from . import agent_info_pb2 as agent__info__pb2
from . import controller_info_pb2 as controller__info__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='messages.proto',
  package='com.rookout',
  syntax='proto3',
  serialized_pb=_b('\n\x0emessages.proto\x12\x0b\x63om.rookout\x1a\rvariant.proto\x1a\x10\x61gent_info.proto\x1a\x15\x63ontroller_info.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"D\n\x0fNewAgentMessage\x12\x31\n\nagent_info\x18\x01 \x01(\x0b\x32\x1d.com.rookout.AgentInformation\">\n\nStackFrame\x12\x10\n\x08\x66ilename\x18\x01 \x01(\t\x12\x0c\n\x04line\x18\x02 \x01(\r\x12\x10\n\x08\x66unction\x18\x03 \x01(\t\"5\n\nStackTrace\x12\'\n\x06\x66rames\x18\x01 \x03(\x0b\x32\x17.com.rookout.StackFrame\"~\n\tException\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x0f\n\x07message\x18\x02 \x01(\t\x12&\n\x08instance\x18\x03 \x01(\x0b\x32\x14.com.rookout.Variant\x12*\n\ttraceback\x18\x04 \x01(\x0b\x32\x17.com.rookout.StackTrace\"\xe2\x03\n\nLogMessage\x12-\n\ttimestamp\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x10\n\x08\x61gent_id\x18\x02 \x01(\t\x12/\n\x05level\x18\x03 \x01(\x0e\x32 .com.rookout.LogMessage.LogLevel\x12\x10\n\x08\x66ilename\x18\x04 \x01(\t\x12\x0c\n\x04line\x18\x05 \x01(\r\x12\x0c\n\x04text\x18\x06 \x01(\t\x12\'\n\targuments\x18\x07 \x03(\x0b\x32\x14.com.rookout.Variant\x12)\n\texception\x18\x08 \x01(\x0b\x32\x16.com.rookout.Exception\x12\x1d\n\x11\x66ormatted_message\x18\t \x01(\tB\x02\x18\x01\x12\x32\n\x10legacy_arguments\x18\n \x01(\x0b\x32\x14.com.rookout.VariantB\x02\x18\x01\x12\x12\n\nclass_name\x18\x0b \x01(\t\x12\x13\n\x0bmethod_name\x18\x0c \x01(\t\x12\x15\n\rcontroller_id\x18\r \x01(\t\"M\n\x08LogLevel\x12\t\n\x05TRACE\x10\x00\x12\t\n\x05\x44\x45\x42UG\x10\x01\x12\x08\n\x04INFO\x10\x02\x12\x0b\n\x07WARNING\x10\x03\x12\t\n\x05\x45RROR\x10\x04\x12\t\n\x05\x46\x41TAL\x10\x05\"\x80\x01\n\x11RuleStatusMessage\x12\x10\n\x08\x61gent_id\x18\x01 \x01(\t\x12\x0f\n\x07rule_id\x18\x02 \x01(\t\x12\x0e\n\x06\x61\x63tive\x18\x03 \x01(\t\x12!\n\x05\x65rror\x18\x04 \x01(\x0b\x32\x12.com.rookout.Error\x12\x15\n\rcontroller_id\x18\x05 \x01(\t\"]\n\x10\x41ugReportMessage\x12\x10\n\x08\x61gent_id\x18\x01 \x01(\t\x12\x0e\n\x06\x61ug_id\x18\x02 \x01(\t\x12\'\n\targuments\x18\x03 \x01(\x0b\x32\x14.com.rookout.Variant\"e\n\x07UserMsg\x12\x31\n\naug_report\x18\x01 \x01(\x0b\x32\x1d.com.rookout.AugReportMessage\x12\x11\n\treport_id\x18\x02 \x01(\t\x12\x14\n\x0cworkspace_id\x18\x03 \x01(\t\"!\n\rAddAugCommand\x12\x10\n\x08\x61ug_json\x18\x01 \x01(\t\"\"\n\x12InitialAugsCommand\x12\x0c\n\x04\x61ugs\x18\x01 \x03(\t\"\"\n\x10RemoveAugCommand\x12\x0e\n\x06\x61ug_id\x18\x01 \x01(\t\"\x15\n\x13InitFinishedMessage\"\x12\n\x10\x43learAugsCommand\"\x1c\n\x0bPingMessage\x12\r\n\x05value\x18\x01 \x01(\x05\"\x1f\n\x0c\x45rrorMessage\x12\x0f\n\x07message\x18\x01 \x01(\t\"e\n\x0e\x43\x32\x43InitMessage\x12:\n\x0e\x63ontrollerInfo\x18\x01 \x01(\x0b\x32\".com.rookout.ControllerInformation\x12\x17\n\x0f\x63onnectedAgents\x18\x02 \x03(\t\" \n\nAgentsList\x12\x12\n\nagents_ids\x18\x01 \x03(\t\":\n\x13\x41gentIdHitCountPair\x12\x10\n\x08\x61gent_id\x18\x01 \x01(\t\x12\x11\n\thit_count\x18\x02 \x01(\r\"f\n\x15HitCountUpdateMessage\x12\x0e\n\x06\x61ug_id\x18\x01 \x01(\t\x12=\n\x13hitCountInformation\x18\x02 \x03(\x0b\x32 .com.rookout.AgentIdHitCountPairb\x06proto3')
  ,
  dependencies=[variant__pb2.DESCRIPTOR,agent__info__pb2.DESCRIPTOR,controller__info__pb2.DESCRIPTOR,google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,])



_LOGMESSAGE_LOGLEVEL = _descriptor.EnumDescriptor(
  name='LogLevel',
  full_name='com.rookout.LogMessage.LogLevel',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='TRACE', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DEBUG', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INFO', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='WARNING', index=3, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ERROR', index=4, number=4,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FATAL', index=5, number=5,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=843,
  serialized_end=920,
)
_sym_db.RegisterEnumDescriptor(_LOGMESSAGE_LOGLEVEL)


_NEWAGENTMESSAGE = _descriptor.Descriptor(
  name='NewAgentMessage',
  full_name='com.rookout.NewAgentMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='agent_info', full_name='com.rookout.NewAgentMessage.agent_info', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=120,
  serialized_end=188,
)


_STACKFRAME = _descriptor.Descriptor(
  name='StackFrame',
  full_name='com.rookout.StackFrame',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='filename', full_name='com.rookout.StackFrame.filename', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='line', full_name='com.rookout.StackFrame.line', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='function', full_name='com.rookout.StackFrame.function', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=190,
  serialized_end=252,
)


_STACKTRACE = _descriptor.Descriptor(
  name='StackTrace',
  full_name='com.rookout.StackTrace',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='frames', full_name='com.rookout.StackTrace.frames', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=254,
  serialized_end=307,
)


_EXCEPTION = _descriptor.Descriptor(
  name='Exception',
  full_name='com.rookout.Exception',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='com.rookout.Exception.type', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='message', full_name='com.rookout.Exception.message', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='instance', full_name='com.rookout.Exception.instance', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='traceback', full_name='com.rookout.Exception.traceback', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=309,
  serialized_end=435,
)


_LOGMESSAGE = _descriptor.Descriptor(
  name='LogMessage',
  full_name='com.rookout.LogMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='com.rookout.LogMessage.timestamp', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='agent_id', full_name='com.rookout.LogMessage.agent_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='level', full_name='com.rookout.LogMessage.level', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='filename', full_name='com.rookout.LogMessage.filename', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='line', full_name='com.rookout.LogMessage.line', index=4,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='text', full_name='com.rookout.LogMessage.text', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='arguments', full_name='com.rookout.LogMessage.arguments', index=6,
      number=7, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='exception', full_name='com.rookout.LogMessage.exception', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='formatted_message', full_name='com.rookout.LogMessage.formatted_message', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\030\001'))),
    _descriptor.FieldDescriptor(
      name='legacy_arguments', full_name='com.rookout.LogMessage.legacy_arguments', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\030\001'))),
    _descriptor.FieldDescriptor(
      name='class_name', full_name='com.rookout.LogMessage.class_name', index=10,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='method_name', full_name='com.rookout.LogMessage.method_name', index=11,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='controller_id', full_name='com.rookout.LogMessage.controller_id', index=12,
      number=13, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _LOGMESSAGE_LOGLEVEL,
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=438,
  serialized_end=920,
)


_RULESTATUSMESSAGE = _descriptor.Descriptor(
  name='RuleStatusMessage',
  full_name='com.rookout.RuleStatusMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='agent_id', full_name='com.rookout.RuleStatusMessage.agent_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='rule_id', full_name='com.rookout.RuleStatusMessage.rule_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='active', full_name='com.rookout.RuleStatusMessage.active', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='error', full_name='com.rookout.RuleStatusMessage.error', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='controller_id', full_name='com.rookout.RuleStatusMessage.controller_id', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=923,
  serialized_end=1051,
)


_AUGREPORTMESSAGE = _descriptor.Descriptor(
  name='AugReportMessage',
  full_name='com.rookout.AugReportMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='agent_id', full_name='com.rookout.AugReportMessage.agent_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='aug_id', full_name='com.rookout.AugReportMessage.aug_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='arguments', full_name='com.rookout.AugReportMessage.arguments', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1053,
  serialized_end=1146,
)


_USERMSG = _descriptor.Descriptor(
  name='UserMsg',
  full_name='com.rookout.UserMsg',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='aug_report', full_name='com.rookout.UserMsg.aug_report', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='report_id', full_name='com.rookout.UserMsg.report_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='workspace_id', full_name='com.rookout.UserMsg.workspace_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1148,
  serialized_end=1249,
)


_ADDAUGCOMMAND = _descriptor.Descriptor(
  name='AddAugCommand',
  full_name='com.rookout.AddAugCommand',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='aug_json', full_name='com.rookout.AddAugCommand.aug_json', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1251,
  serialized_end=1284,
)


_INITIALAUGSCOMMAND = _descriptor.Descriptor(
  name='InitialAugsCommand',
  full_name='com.rookout.InitialAugsCommand',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='augs', full_name='com.rookout.InitialAugsCommand.augs', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1286,
  serialized_end=1320,
)


_REMOVEAUGCOMMAND = _descriptor.Descriptor(
  name='RemoveAugCommand',
  full_name='com.rookout.RemoveAugCommand',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='aug_id', full_name='com.rookout.RemoveAugCommand.aug_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1322,
  serialized_end=1356,
)


_INITFINISHEDMESSAGE = _descriptor.Descriptor(
  name='InitFinishedMessage',
  full_name='com.rookout.InitFinishedMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1358,
  serialized_end=1379,
)


_CLEARAUGSCOMMAND = _descriptor.Descriptor(
  name='ClearAugsCommand',
  full_name='com.rookout.ClearAugsCommand',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1381,
  serialized_end=1399,
)


_PINGMESSAGE = _descriptor.Descriptor(
  name='PingMessage',
  full_name='com.rookout.PingMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='value', full_name='com.rookout.PingMessage.value', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1401,
  serialized_end=1429,
)


_ERRORMESSAGE = _descriptor.Descriptor(
  name='ErrorMessage',
  full_name='com.rookout.ErrorMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='message', full_name='com.rookout.ErrorMessage.message', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1431,
  serialized_end=1462,
)


_C2CINITMESSAGE = _descriptor.Descriptor(
  name='C2CInitMessage',
  full_name='com.rookout.C2CInitMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='controllerInfo', full_name='com.rookout.C2CInitMessage.controllerInfo', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='connectedAgents', full_name='com.rookout.C2CInitMessage.connectedAgents', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1464,
  serialized_end=1565,
)


_AGENTSLIST = _descriptor.Descriptor(
  name='AgentsList',
  full_name='com.rookout.AgentsList',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='agents_ids', full_name='com.rookout.AgentsList.agents_ids', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1567,
  serialized_end=1599,
)


_AGENTIDHITCOUNTPAIR = _descriptor.Descriptor(
  name='AgentIdHitCountPair',
  full_name='com.rookout.AgentIdHitCountPair',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='agent_id', full_name='com.rookout.AgentIdHitCountPair.agent_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='hit_count', full_name='com.rookout.AgentIdHitCountPair.hit_count', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1601,
  serialized_end=1659,
)


_HITCOUNTUPDATEMESSAGE = _descriptor.Descriptor(
  name='HitCountUpdateMessage',
  full_name='com.rookout.HitCountUpdateMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='aug_id', full_name='com.rookout.HitCountUpdateMessage.aug_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='hitCountInformation', full_name='com.rookout.HitCountUpdateMessage.hitCountInformation', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1661,
  serialized_end=1763,
)

_NEWAGENTMESSAGE.fields_by_name['agent_info'].message_type = agent__info__pb2._AGENTINFORMATION
_STACKTRACE.fields_by_name['frames'].message_type = _STACKFRAME
_EXCEPTION.fields_by_name['instance'].message_type = variant__pb2._VARIANT
_EXCEPTION.fields_by_name['traceback'].message_type = _STACKTRACE
_LOGMESSAGE.fields_by_name['timestamp'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LOGMESSAGE.fields_by_name['level'].enum_type = _LOGMESSAGE_LOGLEVEL
_LOGMESSAGE.fields_by_name['arguments'].message_type = variant__pb2._VARIANT
_LOGMESSAGE.fields_by_name['exception'].message_type = _EXCEPTION
_LOGMESSAGE.fields_by_name['legacy_arguments'].message_type = variant__pb2._VARIANT
_LOGMESSAGE_LOGLEVEL.containing_type = _LOGMESSAGE
_RULESTATUSMESSAGE.fields_by_name['error'].message_type = variant__pb2._ERROR
_AUGREPORTMESSAGE.fields_by_name['arguments'].message_type = variant__pb2._VARIANT
_USERMSG.fields_by_name['aug_report'].message_type = _AUGREPORTMESSAGE
_C2CINITMESSAGE.fields_by_name['controllerInfo'].message_type = controller__info__pb2._CONTROLLERINFORMATION
_HITCOUNTUPDATEMESSAGE.fields_by_name['hitCountInformation'].message_type = _AGENTIDHITCOUNTPAIR
DESCRIPTOR.message_types_by_name['NewAgentMessage'] = _NEWAGENTMESSAGE
DESCRIPTOR.message_types_by_name['StackFrame'] = _STACKFRAME
DESCRIPTOR.message_types_by_name['StackTrace'] = _STACKTRACE
DESCRIPTOR.message_types_by_name['Exception'] = _EXCEPTION
DESCRIPTOR.message_types_by_name['LogMessage'] = _LOGMESSAGE
DESCRIPTOR.message_types_by_name['RuleStatusMessage'] = _RULESTATUSMESSAGE
DESCRIPTOR.message_types_by_name['AugReportMessage'] = _AUGREPORTMESSAGE
DESCRIPTOR.message_types_by_name['UserMsg'] = _USERMSG
DESCRIPTOR.message_types_by_name['AddAugCommand'] = _ADDAUGCOMMAND
DESCRIPTOR.message_types_by_name['InitialAugsCommand'] = _INITIALAUGSCOMMAND
DESCRIPTOR.message_types_by_name['RemoveAugCommand'] = _REMOVEAUGCOMMAND
DESCRIPTOR.message_types_by_name['InitFinishedMessage'] = _INITFINISHEDMESSAGE
DESCRIPTOR.message_types_by_name['ClearAugsCommand'] = _CLEARAUGSCOMMAND
DESCRIPTOR.message_types_by_name['PingMessage'] = _PINGMESSAGE
DESCRIPTOR.message_types_by_name['ErrorMessage'] = _ERRORMESSAGE
DESCRIPTOR.message_types_by_name['C2CInitMessage'] = _C2CINITMESSAGE
DESCRIPTOR.message_types_by_name['AgentsList'] = _AGENTSLIST
DESCRIPTOR.message_types_by_name['AgentIdHitCountPair'] = _AGENTIDHITCOUNTPAIR
DESCRIPTOR.message_types_by_name['HitCountUpdateMessage'] = _HITCOUNTUPDATEMESSAGE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

NewAgentMessage = _reflection.GeneratedProtocolMessageType('NewAgentMessage', (_message.Message,), dict(
  DESCRIPTOR = _NEWAGENTMESSAGE,
  __module__ = 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.NewAgentMessage)
  ))
_sym_db.RegisterMessage(NewAgentMessage)

StackFrame = _reflection.GeneratedProtocolMessageType('StackFrame', (_message.Message,), dict(
  DESCRIPTOR = _STACKFRAME,
  __module__ = 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.StackFrame)
  ))
_sym_db.RegisterMessage(StackFrame)

StackTrace = _reflection.GeneratedProtocolMessageType('StackTrace', (_message.Message,), dict(
  DESCRIPTOR = _STACKTRACE,
  __module__ = 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.StackTrace)
  ))
_sym_db.RegisterMessage(StackTrace)

Exception = _reflection.GeneratedProtocolMessageType('Exception', (_message.Message,), dict(
  DESCRIPTOR = _EXCEPTION,
  __module__ = 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.Exception)
  ))
_sym_db.RegisterMessage(Exception)

LogMessage = _reflection.GeneratedProtocolMessageType('LogMessage', (_message.Message,), dict(
  DESCRIPTOR = _LOGMESSAGE,
  __module__ = 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.LogMessage)
  ))
_sym_db.RegisterMessage(LogMessage)

RuleStatusMessage = _reflection.GeneratedProtocolMessageType('RuleStatusMessage', (_message.Message,), dict(
  DESCRIPTOR = _RULESTATUSMESSAGE,
  __module__ = 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.RuleStatusMessage)
  ))
_sym_db.RegisterMessage(RuleStatusMessage)

AugReportMessage = _reflection.GeneratedProtocolMessageType('AugReportMessage', (_message.Message,), dict(
  DESCRIPTOR = _AUGREPORTMESSAGE,
  __module__ = 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.AugReportMessage)
  ))
_sym_db.RegisterMessage(AugReportMessage)

UserMsg = _reflection.GeneratedProtocolMessageType('UserMsg', (_message.Message,), dict(
  DESCRIPTOR = _USERMSG,
  __module__ = 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.UserMsg)
  ))
_sym_db.RegisterMessage(UserMsg)

AddAugCommand = _reflection.GeneratedProtocolMessageType('AddAugCommand', (_message.Message,), dict(
  DESCRIPTOR = _ADDAUGCOMMAND,
  __module__ = 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.AddAugCommand)
  ))
_sym_db.RegisterMessage(AddAugCommand)

InitialAugsCommand = _reflection.GeneratedProtocolMessageType('InitialAugsCommand', (_message.Message,), dict(
  DESCRIPTOR = _INITIALAUGSCOMMAND,
  __module__ = 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.InitialAugsCommand)
  ))
_sym_db.RegisterMessage(InitialAugsCommand)

RemoveAugCommand = _reflection.GeneratedProtocolMessageType('RemoveAugCommand', (_message.Message,), dict(
  DESCRIPTOR = _REMOVEAUGCOMMAND,
  __module__ = 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.RemoveAugCommand)
  ))
_sym_db.RegisterMessage(RemoveAugCommand)

InitFinishedMessage = _reflection.GeneratedProtocolMessageType('InitFinishedMessage', (_message.Message,), dict(
  DESCRIPTOR = _INITFINISHEDMESSAGE,
  __module__ = 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.InitFinishedMessage)
  ))
_sym_db.RegisterMessage(InitFinishedMessage)

ClearAugsCommand = _reflection.GeneratedProtocolMessageType('ClearAugsCommand', (_message.Message,), dict(
  DESCRIPTOR = _CLEARAUGSCOMMAND,
  __module__ = 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.ClearAugsCommand)
  ))
_sym_db.RegisterMessage(ClearAugsCommand)

PingMessage = _reflection.GeneratedProtocolMessageType('PingMessage', (_message.Message,), dict(
  DESCRIPTOR = _PINGMESSAGE,
  __module__ = 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.PingMessage)
  ))
_sym_db.RegisterMessage(PingMessage)

ErrorMessage = _reflection.GeneratedProtocolMessageType('ErrorMessage', (_message.Message,), dict(
  DESCRIPTOR = _ERRORMESSAGE,
  __module__ = 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.ErrorMessage)
  ))
_sym_db.RegisterMessage(ErrorMessage)

C2CInitMessage = _reflection.GeneratedProtocolMessageType('C2CInitMessage', (_message.Message,), dict(
  DESCRIPTOR = _C2CINITMESSAGE,
  __module__ = 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.C2CInitMessage)
  ))
_sym_db.RegisterMessage(C2CInitMessage)

AgentsList = _reflection.GeneratedProtocolMessageType('AgentsList', (_message.Message,), dict(
  DESCRIPTOR = _AGENTSLIST,
  __module__ = 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.AgentsList)
  ))
_sym_db.RegisterMessage(AgentsList)

AgentIdHitCountPair = _reflection.GeneratedProtocolMessageType('AgentIdHitCountPair', (_message.Message,), dict(
  DESCRIPTOR = _AGENTIDHITCOUNTPAIR,
  __module__ = 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.AgentIdHitCountPair)
  ))
_sym_db.RegisterMessage(AgentIdHitCountPair)

HitCountUpdateMessage = _reflection.GeneratedProtocolMessageType('HitCountUpdateMessage', (_message.Message,), dict(
  DESCRIPTOR = _HITCOUNTUPDATEMESSAGE,
  __module__ = 'messages_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.HitCountUpdateMessage)
  ))
_sym_db.RegisterMessage(HitCountUpdateMessage)


_LOGMESSAGE.fields_by_name['formatted_message'].has_options = True
_LOGMESSAGE.fields_by_name['formatted_message']._options = _descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\030\001'))
_LOGMESSAGE.fields_by_name['legacy_arguments'].has_options = True
_LOGMESSAGE.fields_by_name['legacy_arguments']._options = _descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\030\001'))
# @@protoc_insertion_point(module_scope)



import logging

from .backend_logging import add_backend_specific_logs
from .logging_types import LogType


class KWLogger(logging.LoggerAdapter):

    def __init__(self, name, extra_tags=None):
        self.extra_tags = extra_tags or []
        self.warn = self.warning
        return super(KWLogger, self).__init__(name, {})

    def __getattr__(self, item):
        # in case we don't have a function, delagate it to the logger
        return getattr(self.logger, item)

    def process(self, msg, kwargs):
        return msg, self._handle_rookout_kwargs(msg, kwargs)

    def _handle_rookout_kwargs(self, msg, kwargs):
        rookout_extra = {"extraTags": self.extra_tags} if self.extra_tags else {}
        extra = kwargs.get("extra") or {}
        add_backend_specific_logs(kwargs)
        if "rookout_kw_log" in extra:
            # This allow us to have both normal formatted msgs in local mode, and a json version in kubernetes
            kwargs = extra
            del kwargs["rookout_kw_log"]
        for k, v in kwargs.items():
            if k == "exc_info":
                #Same logic as in the original debugger
                continue
            if k == "extra":
                rookout_extra.update(v)
            else:
                field_type = LogType.get_type(k)
                if field_type:
                    self.convert_value_to_type(field_type, k, rookout_extra, v)
                else:
                    rookout_extra["kw_errors"] = rookout_extra.get("kw_errors", []) + [
                        "Missing Kw field: {}".format(k)]

        if not rookout_extra:
            return kwargs

        return {"extra": {"rookout_extra": rookout_extra}}

    def convert_value_to_type(self, field_type, k, rookout_extra, v):
        try:
            rookout_extra[k] = field_type(v)
        except Exception as e:
            rookout_extra["kw_errors"] = rookout_extra.get("kw_errors", []) + [
                "Error convert field {}".format(v)]




import inspect
import six

from .namespace import Namespace
from .container_namespace import ContainerNamespace
from .python_object_namespace import PythonObjectNamespace

from rook.exceptions import RookAttributeNotFound, RookInvalidMethodArguments


from six import string_types

# Fixing a bug in PyPy
import platform
if 'PyPy' == platform.python_implementation():
    import sys
    import six
    FakeFrameType = type(next(six.itervalues(sys._current_frames())))
    from types import FrameType

    def isframe(object):
        return isinstance(object, (FrameType, FakeFrameType))

    inspect.isframe = isframe

class FrameNamespace(Namespace):

    def __init__(self, frame, lineno=None):
        """
        :param lineno: Lineno with which to override the lineno fetched from the frame.
        """
        super(FrameNamespace, self).__init__(self.METHODS)

        self.frame = frame
        if lineno is None:
            self._lineno = inspect.getlineno(frame)
        else:
            self._lineno = lineno
        self._function = self.frame.f_code.co_name

    def read_attribute(self, name):
        if name in self.frame.f_locals:
            return PythonObjectNamespace(self.frame.f_locals[name])
        elif name in self.frame.f_globals:
            return PythonObjectNamespace(self.frame.f_globals[name])
        else:
            raise RookAttributeNotFound(name)

    @property
    def _filename(self):
        return inspect.getsourcefile(self.frame) or inspect.getfile(self.frame)

    def filename(self, args=None):
        return PythonObjectNamespace(self._filename)

    def line(self, args=None):
        return PythonObjectNamespace(self._lineno)

    def function(self, args=None):
        return PythonObjectNamespace(self._function)

    def module(self, args=None):
        module = inspect.getmodule(self.frame)
        if module:
            return PythonObjectNamespace(module.__name__)
        else:
            return PythonObjectNamespace(None)

    def locals(self, args=None):
        depth = None
        dump_config = None

        if args:
            try:
                depth = int(args)
            except ValueError:
                if isinstance(args, string_types):
                    try:
                        dump_config = PythonObjectNamespace.dump_configs[args.lower()]
                    except KeyError:
                        raise RookInvalidMethodArguments('locals()', args)
                else:
                    raise RookInvalidMethodArguments('locals()', args)

        result = {}
        for name, value in six.iteritems(self.frame.f_locals):
            ns = PythonObjectNamespace(value)

            if depth is not None:
                ns.dump_config.max_depth = depth
            elif dump_config is not None:
                ns.dump_config = dump_config

            result[name] = ns

        return ContainerNamespace(result)

    def globals(self, args):
        return PythonObjectNamespace(self.frame.f_globals)

    def dump(self, args):
        return ContainerNamespace({
            'locals': self.locals(args),
            'module': self.module(),
            'filename': self.filename(),
            'line': self.line(),
            'function': self.function(),
        })

    def f_back(self):
        frame = self.frame.f_back

        if frame:
            return FrameNamespace(self.frame.f_back)
        else:
            return PythonObjectNamespace(None)

    def f_next(self):
        frame = self.frame.f_next

        if frame:
            return FrameNamespace(self.frame.f_next)
        else:
            return PythonObjectNamespace(None)

    METHODS = (filename, line, function, module, locals, globals, dump)



import six
from ...logger import logger

from rook.processor.paths.arithmetic_path_internal import parse, ParseError
from rook.exceptions import RookInvalidArithmeticPath, RookNonPrimitiveObjectType, RookAttributeNotFound, RookOperationReadOnly, RookExceptionEvaluationFailed
from ..namespaces.python_object_namespace import PythonObjectNamespace
from ..namespaces.container_namespace import ContainerNamespace


'''
Canopy is used to build and evaluate a tree of operations.
a .peg file (Rookout/canopy/ArithmeticPath.peg) is compiled into the various rooks languages.
Canopy is a PEG parser compiler - and i extend its goal to actually evaluate our arithmetic paths.
'''


ops = {"NE": "!=",
       "=": "==",
       "EQ": "==",
       "LT": "<",
       "GT": ">",
       "GE": ">=",
       "LE": "<=",
       "AND": " and ",
       "OR": " or ",
       "&&": " and ",
       "||": " or "}


level1 = ["*", "/"]
level2 = ["+", "-"]
level3 = ["<=", ">=", "!=", "=", "==", ">", "<", "LT", "GT", "LE", "GE", "EQ", "NE", "lt", "gt", "le", "ge", "eq", "ne"]
level4 = ["in", "IN"]
level5 = ["or", "OR", "||", "and", "AND", "&&"]

allLevels = [level1, level2, level3, level4, level5]

try:
    PRIMITIVE_VALUES = (str, unicode, int, long, float, list)
except NameError:
    PRIMITIVE_VALUES = (str, int, float, list)


class Marker(object):
    pass


class FunctionOperation(Marker):
    def __init__(self, function_name, function_arguments):
        self.function_name = function_name
        self.function_arguments = function_arguments

    def read(self, namespace, create):
        return namespace.call_method(self.function_name, self.function_arguments)

    def write(self, namespace, value):
        raise RookOperationReadOnly(type(self))


class AttributeOperation(Marker):
    def __init__(self, name):
        self.name = name

    def read(self, namespace, create):
        try:
            return namespace.read_attribute(self.name)
        except RookAttributeNotFound as exc:
            if create:
                namespace.write_attribute(self.name, ContainerNamespace({}))
                return namespace.read_attribute(self.name)
            else:
                raise

    def write(self, namespace, value):
        return namespace.write_attribute(self.name, value)


class LookupOperation(Marker):
    def __init__(self, name):
        if name.startswith("'"):
            self.name = name.strip("'")
        elif name.startswith('"'):
            self.name = name.strip('"')
        else:
            self.name = int(name)

    def read(self, namespace, create):
        return namespace.read_key(self.name)

    def write(self, namespace, value):
        raise RookOperationReadOnly(type(self))


class ToolException(Marker):
    def __init__(self, exc):
        self.obj = exc

    def to_string(self):
        return str(self.obj)


class NamespaceResult(Marker):
    def __init__(self, namespace):
        self.obj = namespace
        if isinstance(namespace, PythonObjectNamespace):
            if isinstance(namespace.obj, six.string_types):
                self.text = '\'' + namespace.obj + '\''
            else:
                if namespace.obj is not None:
                    self.text = str(namespace.obj)
                else:
                    self.text = 'None'
        else:
            self.text = 'None'

    def to_string(self):
        return self.text


class NonPrimitiveNamespaceResult(NamespaceResult):
    def __init__(self, namespace, path):
        super(NonPrimitiveNamespaceResult, self).__init__(namespace)
        self.path = path

    def to_string(self):
        return 'NonPrimitiveNamespaceResult'


class Text(Marker):
    def __init__(self, string):
        self.obj = "'" + string + "'"
        self.text = string

    def to_string(self):
        return self.obj


class TextResult(Marker):
    def __init__(self, string):
        self.obj = string
        self.text = string

    def to_string(self):
        return str(self.text)


class Array(Marker):
    def __init__(self, l, string):
        self.obj = l
        self.text = string

    def to_string(self):
        return self.text


class FloatNumber(Marker):
    def __init__(self, num):
        self.obj = float(num)
        self.text = num

    def to_string(self):
        return self.text


class Number(Marker):
    def __init__(self, num):
        self.obj = int(num)
        self.text = num

    def to_string(self):
        return str(self.obj)


class Char(Marker):
    def __init__(self, char):
        self.obj = char
        self.text = "'" + char + "'"

    def to_string(self):
        return self.obj


class Bool(Marker):
    def __init__(self, boolean):
        self.obj = (boolean == 'true') or (boolean == 'True')
        self.text = boolean

    def to_string(self):
        return str(self.obj)


class Undefined(Marker):
    def __init__(self):
        self.obj = None
        self.text = None

    @staticmethod
    def to_string():
        return 'None'


class Opt(Marker):
    def __init__(self, opt):
        self.opt = opt
        self.level = None

        opt_upper_case = opt.upper()

        for key, value in six.iteritems(ops):
            if key == opt_upper_case:
                self.opt = value
                break

        level = 0
        found = False

        for value in allLevels:
            for innerValue in value:
                if opt == innerValue:
                    self.level = level
                    found = True
                    break
            if found:
                break
            level += 1

        if not found:
            #TODO:: throw exception
            pass

    def execute_operation(self, a, b):
        if self.level == 3:
            if not isinstance(b, Array) and not isinstance(b, NamespaceResult):
                return Bool('False')

            if isinstance(b, NamespaceResult):
                for e in b.obj.obj:
                    if str(e) == a.text:
                        return Bool('True')
            else:
                for e in b.obj:
                    if e.text == a.text and e.obj == a.obj:
                        return Bool('True')

            return Bool('False')
        else:
            operation = a.to_string() + self.opt + b.to_string()
            try:
                result = eval(operation)
            except Exception as e:
                raise RookExceptionEvaluationFailed(str(e) + " : " + operation)

            if isinstance(result, bool):
                if not result:
                    if isinstance(a, NonPrimitiveNamespaceResult):
                        if not isinstance(b, Undefined):
                            raise RookNonPrimitiveObjectType(a.path)
                    if isinstance(b, NonPrimitiveNamespaceResult):
                        if not isinstance(a, Undefined):
                            raise RookNonPrimitiveObjectType(b.path)
                return Bool(str(result))

            if isinstance(result, six.integer_types):
                return Number(str(result))

            return TextResult(result)


class Actions(object):
    def __init__(self, namespace):
        self.namespace = None
        self.operations = []
        self.namespace = namespace

    @staticmethod
    def make_lookup_operation(input, start, end, elements):
        return LookupOperation(input[start + 1:end - 1])

    @staticmethod
    def make_function_operation(input, start, end, elements):
        # To build the function name, we will merge the unicode_set and all the unicode_set_with_numbers
        # To build the args we will simply read the atom at index 3
        #   which can be result of another operation; thus checking for exception
        # For further explanation, check ArithmeticPath.peg

        function_name = elements[0].text + elements[1].text

        if isinstance(elements[3], ToolException):
            return elements[3]

        args = elements[3].text
        return FunctionOperation(function_name, args)

    @staticmethod
    def make_function_operation_access(input, start, end, elements):
        # To build the function name, we will merge the unicode_set and all the unicode_set_with_numbers
        # To build the args we will simply read the atom at index 4
        #   which can be result of another operation; thus checking for exception
        # For further explanation, check ArithmeticPath.peg

        function_name = elements[1].text + elements[2].text

        if isinstance(elements[4], ToolException):
            return elements[4]

        args = elements[4].text
        return FunctionOperation(function_name, args)

    @staticmethod
    def make_attribute_operation(input, start, end, elements):
        return AttributeOperation(input[start + 1:end])

    @staticmethod
    def make_attribute(input, start, end, elements):
        return AttributeOperation(input[start:end])

    def make_and_execute_namespace_operation(self, input, start, end, elements):
        # Element 1 will not be null
        # Element 2 is a list of another elements (can be empty)
        # element1.(element2.element3.element4)
        # For further explanation, check ArithmeticPath.peg

        try:
            self.operations = []
            self.operations.append(elements[1])

            for e in elements[2].elements:
                self.operations.append(e)

            # Check if we have some exceptions in the operations chain - might happen if function parsing failed.
            for op in self.operations:
                if isinstance(op, ToolException):
                    return op

            result = elements[1].read(self.namespace, False)
            for e in elements[2].elements:
                result = e.read(result, False)

            if isinstance(result, PythonObjectNamespace) and result.obj and not isinstance(result.obj, PRIMITIVE_VALUES):
                # returning this class that mark its non primitive for additional information
                # this will help us in case of non True result when comparing this object
                return NonPrimitiveNamespaceResult(result, input[start:end])

            return NamespaceResult(result)
        except Exception as e:
            return ToolException(e)

    @staticmethod
    def make_comp_exp_ex(input, start, end, elements):
        # Element 2 is the actual expression
        # For further explanation, check ArithmeticPath.peg
        return elements[2]

    @staticmethod
    def make_comp_exp(input, start, end, elements):
        # We can assume the following: atom ( opt_ atom )*
        # the first (which must be) will be simple atom
        # the second and so forth will always be pair <Opt, Atom>
        # Its important to remember that this execution will handle the inner brackets if exist
        # In order to handle priority (which i could not figure out if available with canopy library):
        # 1. lets make the tree flat
        # 2. handle priority ourselves - (atom opt1 atom) will be handled before (atom opt2 atom)
        #   and will return TreeNode with result
        # For further explanation, check ArithmeticPath.peg

        # handle case the size is 1
        if len(elements[1].elements) == 0:
            return elements[0]

        if isinstance(elements[0], ToolException):
            return elements[0]

        flat_elements = [elements[0]]
        for n in elements[1].elements:
            flat_elements.append(n.elements[0])

            if isinstance(n.elements[1], ToolException):
                return n.elements[1]
            flat_elements.append(n.elements[1])

        while len(flat_elements) > 1:
            stop = False
            for level in range(len(allLevels)):
                for i in range(1, len(flat_elements), 2):
                    current_opt = flat_elements[i]
                    if current_opt.level == level:
                        result = current_opt.execute_operation(flat_elements[i - 1], flat_elements[i + 1])
                        flat_elements = flat_elements[:i - 1] + [result] + flat_elements[i + 2:]
                        stop = True
                        break
                if stop:
                    break

        return flat_elements[0]

    @staticmethod
    def make_opt(input, start, end, elements):
        return Opt(elements[1].text)

    @staticmethod
    def make_apostrophe_string(input, start, end, elements):
        return Text(elements[2].text)

    @staticmethod
    def make_string(input, start, end, elements):
        return Text(elements[2].text)

    @staticmethod
    def make_list(input, start, end, elements):
        tree_flatter = TreeFlatter()
        result = tree_flatter.flat_tree(elements[3])
        return Array(result, input[start:end])

    @staticmethod
    def make_float(input, start, end, elements):
        return FloatNumber(input[start:end].replace(' ', ''))

    @staticmethod
    def make_number(input, start, end, elements):
        return Number(input[start:end].replace(' ', ''))

    @staticmethod
    def make_char(input, start, end, elements):
        return Char(input[start:end].replace(' ', ''))

    @staticmethod
    def make_bool(input, start, end, elements):
        return Bool(input[start:end].replace(' ', ''))

    @staticmethod
    def make_null(input, start, end, elements):
        return Undefined()

    @staticmethod
    def make_undefined(input, start, end, elements):
        return Undefined()


class TreeFlatter(object):
    def __init__(self):
        self.flatted_tree = []

    def flat_tree(self, element):
        if isinstance(element, Marker):
            self.flatted_tree.append(element)

        for e in element.elements:
            if isinstance(e, Marker):
                self.flatted_tree.append(e)
            else:
                self.flat_tree(e)

        return self.flatted_tree


class ArithmeticPath(object):
    NAME = 'calc'

    def __init__(self, configuration, factory=None):
        if isinstance(configuration, six.string_types):
            string = configuration
        else:
            string = configuration['path']

        self.expression = string
        if self.expression == "":
            raise RookInvalidArithmeticPath(self.expression)

    def read_from(self, root_namespace):
        context = Actions(root_namespace)

        try:
            result = parse(self.expression, actions=context)
        except ParseError as e:
            raise RookInvalidArithmeticPath(self.expression, e)
        return self.normalize_result(result)

    def write_to(self, root_namespace, value):
        context = Actions(root_namespace)

        try:
            # initialize operations chain - by parsing the expression - ignore return value
            parse(self.expression, actions=context)
        except ParseError as e:
            raise RookInvalidArithmeticPath(self.expression, e)

        # execute the operation chain.
        size = len(context.operations)
        if size == 0:
            raise RookInvalidArithmeticPath(self.expression)

        i = 0
        namespace = root_namespace
        for _ in range(i, size - 1):
            namespace = context.operations[i].read(namespace, True)
            i += 1

        context.operations[i].write(namespace, value)

    @staticmethod
    def normalize_result(result):
        if isinstance(result, Array):
            new_arr = []
            for e in result.obj:
                new_arr.append(e.obj)
            return PythonObjectNamespace(new_arr)

        if isinstance(result, NamespaceResult):
            return result.obj

        if isinstance(result, ToolException):
            raise result.obj

        if isinstance(result, NonPrimitiveNamespaceResult):
            raise result.obj

        return PythonObjectNamespace(result.obj)



import logging
import six
import inspect
import os
import threading

from rook.logger import logger

from rook.processor.namespaces.container_namespace import ContainerNamespace
from rook.processor.namespaces.frame_namespace import FrameNamespace
from rook.augs.processor_extensions.namespaces.log_record_namespace import LogRecordNamespace


class LoggingLocationService(object):

    NAME = "logging"

    class Handler(logging.Handler):

        def __init__(self, logger):
            super(LoggingLocationService.Handler, self).__init__()

            self._name = logger.name

            self._logger = logger
            self._old_remove_handler = self._logger.removeHandler

            # Do not allow us to be removed
            def removeHandler(handler):
                if handler != self:
                    self._old_remove_handler(handler)
            self._logger.removeHandler = removeHandler

            self._logger.addHandler(self)

            self._augs = {}
            self._lock = threading.RLock()

        def close(self):
            self._logger.removeHandler = self._old_remove_handler
            self._logger.removeHandler(self)

        def emit(self, record):
            try:
                frame = FrameNamespace(self._get_frame(record))
                extracted = ContainerNamespace({'log_record': LogRecordNamespace(record)})

                with self._lock:
                    for aug in six.itervalues(self._augs):
                        aug.execute(frame, extracted)
            except:
                logger.exception("Error while processing log record")

        def add_aug(self, aug):
            with self._lock:
                self._augs[aug.aug_id] = aug

                aug.set_active()

        def remove_aug(self, aug_id):
            with self._lock:
                try:
                    aug = self._augs[aug_id]
                except KeyError:
                    return

                del self._augs[aug_id]
                aug.set_removed()

        def clear_augs(self):
            with self._lock:
                aug_ids = list(self._augs.keys())

                for aug_id in aug_ids:
                    self.remove_aug(aug_id)

        def empty(self):
            return len(self._augs) == 0

        def _get_frame(self, record):
            # Skip the top two frames (ours)
            frame = inspect.currentframe().f_back.f_back

            while hasattr(frame, "f_code"):
                filename = os.path.normcase(frame.f_code.co_filename)
                if filename == logging._srcfile:
                    frame = frame.f_back
                else:
                    return frame

            return None

    def __init__(self):
        self._hooked = False

        self._handlers = dict()
        self.rootHandler = None
        self._lock = threading.RLock()

    def add_logging_aug(self, name, aug):
        with self._lock:
            self._installHooks()

            try:
                handler = self._handlers[name]
            except KeyError:
                logger = logging.getLogger(name)
                handler = self.Handler(logger)
                self._handlers[name] = handler

                # We document root handler for basicConfig support
                if logger == logging.root:
                    self.rootHandler = handler

        handler.add_aug(aug)

    def remove_aug(self, aug_id):
        with self._lock:
            for handler in six.itervalues(self._handlers):
                handler.remove_aug(aug_id)

        self.prune_handlers()

    def clear_augs(self):
        with self._lock:
            for handler in six.itervalues(self._handlers):
                handler.clear_augs()

        self.prune_handlers()

    def prune_handlers(self):
        with self._lock:
            loggers_to_prune = [logger for logger, handler in six.iteritems(self._handlers) if handler.empty()]

            for logger in loggers_to_prune:

                self._handlers[logger].close()

                # If we just closed rootHandler, remove it
                if self._handlers[logger] == self.rootHandler:
                    self.rootHandler = None

                del self._handlers[logger]

    def close(self):
        self.clear_augs()

    def _installHooks(self):
        with self._lock:
            if self._hooked:
                return

            self.originalBasicConfig = logging.basicConfig
            logging.basicConfig = self._basicConfigHook

            for name in ["critical", "error", "warning", "info", "debug", "log"]:
                self._wrap_logging_method(name)

            self._hooked = True

    def _basicConfigHook(self, **kwargs):
        logging._acquireLock()

        try:
            if self.rootHandler:
                self.rootHandler._old_remove_handler(self.rootHandler)

                try:
                    self.originalBasicConfig(**kwargs)
                finally:
                    logging.root.addHandler(self.rootHandler)

            else:
                self.originalBasicConfig(**kwargs)
        finally:
            logging._releaseLock()

    def _isBasicConfigNeeded(self):
        return len(logging.root.handlers) == 0 or (len(logging.root.handlers) == 1 and self.rootHandler)

    # based off https://github.com/getsentry/raven-python/blob/master/raven/breadcrumbs.py
    def _wrap_logging_method(self, name):

        func = logging.__dict__[name]
        code = func.__code__

        logging_srcfile = logging._srcfile
        if logging_srcfile is None:
            logging_srcfile = os.path.normpath(
                logging.currentframe.__code__.co_filename
            )

        # This requires a bit of explanation why we're doing this.  Due to how
        # logging itself works we need to pretend that the method actually was
        # created within the logging module.  There are a few ways to detect
        # this and we fake all of them: we use the same function globals (the
        # one from the logging module), we create it entirely there which
        # means that also the filename is set correctly.  This fools the
        # detection code in logging and it makes logging itself skip past our
        # code when determining the code location.
        #
        # Because we point the globals to the logging module we now need to
        # refer to our own functions (original and the crumb recording
        # function) through a closure instead of the global scope.
        #
        # We also add a lot of newlines in front of the code so that the
        # code location lines up again in case someone runs inspect.getsource
        # on the function.
        ns = {}
        eval(compile('''%(offset)sif 1:
        def factory(name, isBasicConfigNeeded):
            def %(name)s(*args, **kwargs):
                if isBasicConfigNeeded():
                    basicConfig()
                
                return getattr(root, name)(*args, **kwargs)
            return %(name)s
        \n''' % {
            'offset': '\n' * (code.co_firstlineno - 3),
            'name': name,
        }, logging_srcfile, 'exec'), logging.__dict__, ns)

        new_func = ns['factory'](name, self._isBasicConfigNeeded)
        new_func.__doc__ = func.__doc__

        assert code.co_firstlineno == code.co_firstlineno

        # In theory this should already be set correctly, but in some cases
        # it is not.  So override it.
        new_func.__module__ = func.__module__
        new_func.__name__ = func.__name__

        logging.__dict__[name] = new_func




import json
import logging
from logging.handlers import RotatingFileHandler

from six import iteritems


class RookoutLogHandler(logging.Handler):

    @staticmethod
    def update_record(record):
        # Enable override of basic log elements
        def update_from_override(original_name, update_name):
            update_data = getattr(record, update_name, None)
            if update_data is not None:
                setattr(record, original_name, update_data)

        updates = {"_module": "module",
                   "_funcName": "funcName",
                   "_lineno": "lineno"}

        for update, original in iteritems(updates):
            update_from_override(original, update)
        rookout_args = ""
        if hasattr(record, "rookout_extra"):
            rookout_args = json.dumps(getattr(record, "rookout_extra"))
        setattr(record, "rookout_extra", rookout_args)
        return record

    def emit(self, record):
        record = self.update_record(record)
        # Pass to actual emit
        super(RookoutLogHandler, self).emit(record)


class RookoutStreamHandler(RookoutLogHandler, logging.StreamHandler):
    pass


class RookoutFileHandler(RookoutLogHandler, RotatingFileHandler):
    pass



# -*- coding: utf-8 -*-
"""
Copied from
https://raw.githubusercontent.com/celery/celery/master/celery/contrib/rdb.py
Commit 9d345f583631709cfdd38b77d2947ec74d83a562

License (https://github.com/celery/celery/blob/master/LICENSE):
Copyright (c) 2015-2016 Ask Solem & contributors.  All rights reserved.
Copyright (c) 2012-2014 GoPivotal, Inc.  All rights reserved.
Copyright (c) 2009, 2010, 2011, 2012 Ask Solem, and individual contributors.  All rights reserved.

Celery is licensed under The BSD License (3 Clause, also known as
the new BSD license).  The license is an OSI approved Open Source
license and is GPL-compatible(1).

The license text can also be found here:
http://www.opensource.org/licenses/BSD-3-Clause

License
=======

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Ask Solem, nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Ask Solem OR CONTRIBUTORS
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

Documentation License
=====================

The documentation portion of Celery (the rendered contents of the
"docs" directory of a software distribution or checkout) is supplied
under the "Creative Commons Attribution-ShareAlike 4.0
International" (CC BY-SA 4.0) License as described by
https://creativecommons.org/licenses/by-sa/4.0/

Footnotes
=========
(1) A GPL-compatible license makes it possible to
    combine Celery with other software that is released
    under the GPL, it does not mean that we're distributing
    Celery under the GPL license.  The BSD license, unlike the GPL,
    let you distribute a modified version without making your
    changes open source.

########################################################################################################################
Remote Debugger.

Introduction
============

This is a remote debugger for Celery tasks running in multiprocessing
pool workers.  Inspired by a lost post on dzone.com.

Usage
-----

.. code-block:: python

    from celery.contrib import rdb
    from celery import task

    @task()
    def add(x, y):
        result = x + y
        rdb.set_trace()
        return result

Environment Variables
=====================

.. envvar:: CELERY_RDB_HOST

``CELERY_RDB_HOST``
-------------------

    Hostname to bind to.  Default is '127.0.01' (only accessable from
    localhost).

.. envvar:: CELERY_RDB_PORT

``CELERY_RDB_PORT``
-------------------

    Base port to bind to.  Default is 6899.
    The debugger will try to find an available port starting from the
    base port.  The selected port will be logged by the worker.
"""
from __future__ import absolute_import, print_function, unicode_literals
import errno
import os
import socket
import sys
from pdb import Pdb
from billiard.process import current_process
from celery.five import range

__all__ = [
    'CELERY_RDB_HOST', 'CELERY_RDB_PORT', 'DEFAULT_PORT',
    'Rdb', 'debugger', 'set_trace',
]

DEFAULT_PORT = 6899

CELERY_RDB_HOST = os.environ.get('CELERY_RDB_HOST') or '127.0.0.1'
CELERY_RDB_PORT = int(os.environ.get('CELERY_RDB_PORT') or DEFAULT_PORT)

#: Holds the currently active debugger.
_current = [None]

_frame = getattr(sys, '_getframe')

NO_AVAILABLE_PORT = """\
{self.ident}: Couldn't find an available port.

Please specify one using the CELERY_RDB_PORT environment variable.
"""

BANNER = """\
{self.ident}: Ready to connect: telnet {self.host} {self.port}

Type `exit` in session to continue.

{self.ident}: Waiting for client...
"""

SESSION_STARTED = '{self.ident}: Now in session with {self.remote_addr}.'
SESSION_ENDED = '{self.ident}: Session with {self.remote_addr} ended.'


class Rdb(Pdb):
    """Remote debugger."""

    me = 'Remote Debugger'
    _prev_outs = None
    _sock = None

    def __init__(self, host=CELERY_RDB_HOST, port=CELERY_RDB_PORT,
                 port_search_limit=100, port_skew=+0, out=sys.stdout):
        self.active = True
        self.out = out

        self._prev_handles = sys.stdin, sys.stdout

        self._sock, this_port = self.get_avail_port(
            host, port, port_search_limit, port_skew,
        )
        self._sock.setblocking(1)
        self._sock.listen(1)
        self.ident = '{0}:{1}'.format(self.me, this_port)
        self.host = host
        self.port = this_port
        self.say(BANNER.format(self=self))

    def connect(self):
        self._client, address = self._sock.accept()
        self._client.setblocking(1)
        self.remote_addr = ':'.join(str(v) for v in address)
        self.say(SESSION_STARTED.format(self=self))
        self._handle = sys.stdin = sys.stdout = self._client.makefile('rw')
        Pdb.__init__(self, completekey='tab',
                     stdin=self._handle, stdout=self._handle)

    def get_avail_port(self, host, port, search_limit=100, skew=+0):
        try:
            _, skew = current_process().name.split('-')
            skew = int(skew)
        except ValueError:
            pass
        this_port = None
        for i in range(search_limit):
            _sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            _sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            this_port = port + skew + i
            try:
                _sock.bind((host, this_port))
            except socket.error as exc:
                if exc.errno in [errno.EADDRINUSE, errno.EINVAL]:
                    continue
                raise
            else:
                return _sock, this_port
        else:
            raise Exception(NO_AVAILABLE_PORT.format(self=self))

    def say(self, m):
        print(m, file=self.out)

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self._close_session()

    def _close_session(self):
        self.stdin, self.stdout = sys.stdin, sys.stdout = self._prev_handles
        if self.active:
            if self._handle is not None:
                self._handle.close()
            if self._client is not None:
                self._client.close()
            if self._sock is not None:
                self._sock.close()
            self.active = False
            self.say(SESSION_ENDED.format(self=self))

    def do_continue(self, arg):
        self._close_session()
        self.set_continue()
        return 1
    do_c = do_cont = do_continue

    def do_quit(self, arg):
        self._close_session()
        self.set_quit()
        return 1
    do_q = do_exit = do_quit

    def set_quit(self):
        # this raises a BdbQuit exception that we're unable to catch.
        sys.settrace(None)


def debugger():
    """Return the current debugger instance, or create if none."""
    rdb = _current[0]
    if rdb is None or not rdb.active:
        rdb = _current[0] = Rdb()
    return rdb


def set_trace(frame=None):
    """Set break-point at current location, or a specified frame."""
    if frame is None:
        frame = _frame().f_back
    return debugger().set_trace(frame)



from rook.logger import logger
from rook.processor.error import Error

from .aug_factory import AugFactory


class AugsManager(object):
    def __init__(self, trigger_services, output):
        self._trigger_services = trigger_services
        self._output = output
        self._aug_factory = AugFactory(output)
        self._augs = set()

    def initialize_augs(self, augs):
        leftovers = self._augs.copy()

        for aug in augs:
            try:
                aug_id = aug['id']
            except Exception:
                continue

            if aug_id in leftovers:
                leftovers.remove(aug_id)
                continue

            self.add_aug(aug)

        for leftover in leftovers:
            self.remove_aug(leftover)

    def add_aug(self, configuration):
        try:
            aug = self._aug_factory.get_aug(configuration)
        except Exception as exc:
            message = "Failed to parse aug"
            logger.exception(message)

            try:
                aug_id = configuration['id']
            except Exception:
                return

            self._output.send_rule_status(aug_id, "Error", Error(exc=exc, message=message))
            return

        if aug.aug_id in self._augs:
            logger.debug('Aug already set - %s', aug.aug_id)
            return

        logger.debug("Adding aug-\t%s", aug.aug_id)

        aug.add_aug(self._trigger_services)
        self._augs.add(aug.aug_id)

    def remove_aug(self, aug_id):
        logger.debug("Removing aug-\t%s", aug_id)

        self._trigger_services.remove_aug(aug_id)
        self._augs.remove(aug_id)

    def clear_augs(self):
        logger.debug("Clearing all augs")

        # Explicitly delete (and report) all augs we know about)
        for aug in self._augs.copy():
            self.remove_aug(aug)

        # Just in case- delete anything that may have been left behind
        self._trigger_services.clear_augs()



import socket
import atexit
import uuid
import threading
import time
from six.moves.queue import Queue, Empty
from six.moves.urllib.parse import urlparse
import select
import errno
import certifi
import ssl

import rook.com_ws.socketpair_compat  # do not remove - adds socket.socketpair on Windows
import websocket

from rook.com_ws import information
from rook.logger import logger
import rook.protobuf2.messages_pb2 as messages_pb
import rook.protobuf2.envelope_pb2 as envelope_pb
from rook.config import AgentComConfiguration, VersionConfiguration
from rook.exceptions import RookCommunicationException, RookInvalidToken


def wrap_in_envelope(message):
    envelope = envelope_pb.Envelope()
    envelope.timestamp.GetCurrentTime()
    envelope.msg.Pack(message)

    return envelope


class ExitThreadSentinel(object):
    pass


class FlushMessagesEvent(object):
    def __init__(self):
        self.event = threading.Event()


class MessageCallback(object):
    def __init__(self, cb, persistent):
        self.cb = cb
        self.persistent = persistent


class AgentCom(object):
    def __init__(self, host, port, proxy, token, labels, tags):
        self._thread = threading.Thread(name="rookout-" + type(self).__name__, target=self._connect)
        self._thread.daemon = True

        self._host = host
        self._port = port
        self._proxy = proxy
        self._token = token
        self._labels = labels or {}
        self._tags = tags or []
        self._reset_id()

        self._loop = None
        self._connection = None
        self._queue = Queue()

        self._running = True

        self._ready_event = threading.Event()
        self._connection_error = None

        self._callbacks = {}

        def set_ready_event(*args):
            self._ready_event.set()

        self.once('InitialAugsCommand', set_ready_event)

        atexit.register(self.stop)

    def start(self):
        self._thread.start()

    def stop(self):
        self._running = False

        if self._connection is not None:
            self._connection.close(1000)

    def add(self, message):
        self._queue.put(wrap_in_envelope(message))

    def on(self, message_name, callback):
        self._register_callback(message_name, MessageCallback(callback, True))

    def once(self, message_name, callback):
        self._register_callback(message_name, MessageCallback(callback, False))

    def await_message(self, message_name):
        event = threading.Event()
        self.once(message_name, lambda _: event.set())

        return event

    def wait_for_ready(self, timeout=None):
        if not self._ready_event.wait(timeout):
            raise RookCommunicationException()
        else:
            if self._connection_error is not None:
                raise self._connection_error

    def _connect(self):
        retry = 0
        backoff = AgentComConfiguration.BACK_OFF
        connected = False
        last_successful_connection = 0

        while self._running:
            try:
                try:
                    if connected and time.time() >= last_successful_connection + AgentComConfiguration.RESET_BACKOFF_TIMEOUT:
                        retry = 0
                        backoff = AgentComConfiguration.BACK_OFF

                    self._connection = self._create_connection()

                    self._register_agent()

                except websocket.WebSocketBadStatusException as e:
                    if e.status_code == 403:  # invalid token
                        self._connection_error = RookInvalidToken(self._token)
                        self._ready_event.set()
                        return
                    raise
            except Exception as e:
                retry += 1
                backoff = min(backoff * 2, AgentComConfiguration.MAX_SLEEP)
                connected = False

                if hasattr(e, 'message') and e.message:
                    reason = e.message
                else:
                    reason = str(e)

                logger.info('Connection failed; reason = %s, retry = #%d, waiting %.3fs', reason, retry, backoff)

                time.sleep(backoff)
                continue
            else:
                connected = True
                last_successful_connection = time.time()

            logger.debug("WebSocket connected successfully")

            stop_socket_incoming, incoming_client_socket = socket.socketpair()
            outgoing_exit_sentinel = ExitThreadSentinel()

            def signal_stop_incoming_thread():
                logger.debug("calling stop incoming thread")
                try:
                    stop_socket_incoming.send(b'1')
                except socket.error as socket_error:
                    if socket_error.errno == errno.EPIPE:
                        return
                    logger.exception("Failed to signal stop")
                except:
                    pass
                finally:
                    stop_socket_incoming.close()

            def signal_stop_outgoing_thread():
                logger.debug("calling stop outgoing thread")
                try:
                    incoming_client_socket.close()
                except:
                    pass

                try:
                    self._queue.put(outgoing_exit_sentinel)
                except Exception:
                    logger.exception("Failed to write to queue")
                    pass

            routines = [threading.Thread(name="rookout-incoming-thread", target=self._incoming, args=(incoming_client_socket, signal_stop_outgoing_thread)),
                        threading.Thread(name="rookout-outgoing-thread", target=self._outgoing, args=(outgoing_exit_sentinel, signal_stop_incoming_thread))]

            got_initial_augs = self.await_message('InitialAugsCommand')

            for routine in routines:
                routine.daemon = True
                routine.start()

            if got_initial_augs.wait(AgentComConfiguration.TIMEOUT):
                logger.info("Finished initialization")
            else:
                logger.warning("Timeout waiting for initial augs")

            for routine in routines:
                routine.join()

            logger.debug("Reconnecting")

    def _incoming(self, stop_socket, on_exit):
        try:
            while True:
                try:
                    rlist, _, xlist = select.select([self._connection, stop_socket], [], [self._connection, stop_socket], AgentComConfiguration.PING_TIMEOUT)
                    if stop_socket in rlist or stop_socket in xlist or self._connection in xlist:
                        logger.debug("Incoming thread signaled to stop")
                        break  # signaled to stop

                    if len(rlist) == 0 and len(xlist) == 0:
                        logger.debug("Incoming thread - ping timeout")
                        break  # select timed out - ping timeout

                    # it wasn't stop_socket, so it's self._connection -> we can read. we have to read including
                    # control frames, otherwise select might return and recv() wouldn't actually return anything
                    code, msg = self._connection.recv_data(control_frame=True)
                    if code != websocket.ABNF.OPCODE_BINARY:
                        continue

                    if msg is None:
                        # socket disconnected
                        logger.debug("Incoming thread - socket disconnected")
                        break

                    envelope = envelope_pb.Envelope()
                    envelope.ParseFromString(msg)
                    self._handle_incoming_message(envelope)
                except (socket.error, websocket.WebSocketConnectionClosedException):
                    logger.debug("Incoming thread - socket disconnected")
                    break
        except:
            logger.exception("Incoming thread failed")

        finally:
            logger.debug("Incoming thread exiting")
            on_exit()

    def flush_all_messages(self):
        flush_event = FlushMessagesEvent()
        self._queue.put(flush_event)
        flush_event.event.wait(AgentComConfiguration.FLUSH_TIMEOUT)

    def _outgoing(self, outgoing_exit_sentinel, on_exit):
        try:
            last_ping = time.time()
            self._connection.ping()

            while True:
                msg = None
                if (time.time() - last_ping) >= AgentComConfiguration.PING_INTERVAL:
                    last_ping = time.time()
                    self._connection.ping()

                try:
                    msg = self._queue.get(timeout=AgentComConfiguration.PING_INTERVAL)
                    if isinstance(msg, ExitThreadSentinel):
                        if msg is outgoing_exit_sentinel:
                            break
                        continue  # if it's an ExitThreadSentinel but not from this specific thread, just skip it
                    if isinstance(msg, FlushMessagesEvent):
                        msg.event.set()
                        continue
                    self._send(msg)
                except Empty:
                    continue
                except (socket.error, websocket.WebSocketConnectionClosedException):
                    if msg is not None:
                        self._queue.put(msg)
                    break
        except:
            logger.exception("Outgoing thread failed")

        finally:
            logger.debug("Outgoing thread exiting")
            on_exit()

    def _create_connection(self):
        url = '{}:{}/v1'.format(self._host, self._port)
        headers = {
            'User-Agent': 'RookoutAgent/{}+{}'.format(VersionConfiguration.VERSION, VersionConfiguration.COMMIT)
        }

        if self._token is not None:
            headers["X-Rookout-Token"] = self._token

        proxy_host, proxy_port = self._get_proxy()

        connect_args = (url,)
        connect_kwargs = dict(header=headers,
                              timeout=AgentComConfiguration.TIMEOUT,
                              http_proxy_host=proxy_host, http_proxy_port=proxy_port)

        try:
            # connect using system certificates
            conn = websocket.create_connection(*connect_args, **connect_kwargs)
        except (ssl.SSLError, ssl.CertificateError):
            # connect using certifi certificate bundle
            # (Python 2.7.15+ from python.org on macOS rejects our CA, see RK-3383)
            connect_kwargs['sslopt'] = dict(ca_certs=certifi.where())
            logger.debug("Got SSL error when connecting using system CA cert store, falling back to certifi")
            conn = websocket.create_connection(*connect_args, **connect_kwargs)
        conn.settimeout(None)
        return conn

    def _get_proxy(self):
        if self._proxy is None:
            return None, None

        try:
            if not self._proxy.startswith("http://"):
                self._proxy = "http://" + self._proxy

            url = urlparse(self._proxy, "http://")

            logger.debug("Connecting via proxy: %s", url.netloc)

            return url.hostname, url.port
        except ValueError:
            return None, None

    def _register_agent(self):
        logger.info('Registering agent with id %s', self.id)
        info = information.collect()
        info.agent_id = self.id
        info.labels = self._labels
        info.tags = self._tags

        m = messages_pb.NewAgentMessage()
        m.agent_info.CopyFrom(information.pack_agent_info(info))

        return self._send(wrap_in_envelope(m))

    AcceptedMessageTypes = [
        messages_pb.InitialAugsCommand,
        messages_pb.AddAugCommand,
        messages_pb.ClearAugsCommand,
        messages_pb.PingMessage,
        messages_pb.RemoveAugCommand
    ]

    def _handle_incoming_message(self, envelope):
        for message_type in self.AcceptedMessageTypes:
            if envelope.msg.Is(message_type.DESCRIPTOR):
                message = message_type()
                envelope.msg.Unpack(message)
                type_name = message.DESCRIPTOR.name

                callbacks = self._callbacks.get(type_name)

                if callbacks:
                    persistent_callbacks = []

                    # Trigger all persistent callbacks first
                    for callback in callbacks:
                        try:
                            if callback.persistent:
                                callback.cb(message)
                        except:
                            pass
                        finally:
                            if callback.persistent:
                                persistent_callbacks.append(callback)

                    # Trigger all non persistent callbacks
                    for callback in callbacks:
                        try:
                            if not callback.persistent:
                                callback.cb(message)
                        except:
                            pass

                    self._callbacks[type_name] = persistent_callbacks

    def _send(self, message):
        return self._connection.send_binary(message.SerializeToString())

    def _register_callback(self, message_name, callback):
        self._callbacks.setdefault(message_name, []).append(callback)

    def _reset_id(self):
        self.id = uuid.uuid4().hex



from string import Formatter

from rook.logger import logger
from ..namespaces.formatted_namespace import FormattedNamespace
from rook.exceptions import RookUnknownObject
from ..error import Error
from ...user_warnings import UserWarnings

class Format(object):
    class NamespaceFormatter(Formatter):
        def __init__(self, namespace, factory, *args, **kwargs):
            super(Format.NamespaceFormatter, self).__init__(*args, **kwargs)

            self.namespace = namespace
            self.factory = factory

        def get_field(self, field_name, args, kwargs):
            try:
                path = self.factory.get_path(field_name)
                value = path.read_from(self.namespace)
                return value.obj, False
            except Exception:
                logger.exception('Failed to read value- %s', field_name)
                UserWarnings.send_warning(Error(RookUnknownObject(field_name)))
                return '<Missing>', False

    def __init__(self, arguments, factory):
        self.factory = factory

        self._format = arguments['format']
        self._target = factory.get_path(arguments['path'])

        self._paths = list()
        if 'values' in arguments:
            for item in arguments['values']:
                self._paths.append(factory.get_path(item))

    def execute(self, namespace):
        values = []
        for path in self._paths:
            values.append(path.read_from(namespace).obj)

        if values:
            formatter = Formatter()
        else:
            formatter = self.NamespaceFormatter(namespace, self.factory)

        message = formatter.vformat(self._format, values, {})

        self._target.write_to(namespace, FormattedNamespace(message))
